# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.view import CommonView

# GObject
from gi.repository import Gtk

# Iteration
from itertools import product

# Unittest
import unittest


# ------------------------------------------------------------------------------
#   Tests
# ------------------------------------------------------------------------------

class DummyGeodeGtkTreeView(CommonView, Gtk.TreeView):

    def __init__(self, model, *args, **kwargs):
        """ Constructor
        """

        CommonView.__init__(self, Gtk.TreeView, model, **kwargs)

        self.set_model(self.inner_model)


class GeodeGEMWidgetsViewTC(unittest.TestCase):

    def check_converted_treeiter(self, view, treeiter, index):
        """ Check converted treeiter to ensure this is the same row

        Parameters
        ----------
        view : geode_gem.widgets.view.CommonView
            Parent view for the specified treeiter
        treeiter : Gtk.TreeIter
            Treeiter to compare
        index : number
            Treeiter value to compare with converted treeiter data
        """

        model = view.get_model()

        converted_treeiter = view.convert_child_iter_to_iter(treeiter)
        self.assertIsNotNone(converted_treeiter)
        self.assertEqual(model.get_value(converted_treeiter, 1),
                         view.list_model.get_value(treeiter, 1))
        self.assertEqual(model.get_value(converted_treeiter, 1),
                         f"index-{index}")

    def generate_treeview_and_treeiters(self, number=10, **kwargs):
        """ Generate both treeview and treeiters

        Parameters
        ----------
        number : int, optional
            Treeview rows number to generate
        kwargs : dict, optional
            Additionnal positional parameters for GeodeGtkTreeView

        Returns
        -------
        geode_gem.widgets.view.GeodeGtkTreeView, list
            Generated treeview and treeiters list
        """

        view = DummyGeodeGtkTreeView(Gtk.ListStore(int, str), **kwargs)

        treeiters = list()
        for index in range(number):
            treeiters.append(view.append([index, f"index-{index}"]))
            self.assertIsInstance(treeiters[index], Gtk.TreeIter)

        return view, treeiters

    def generate_truth_table(self, dimension=2):
        """ Generate a truth table list from a specific dimension

        Parameters
        ----------
        dimension : int, optional
            Table dimension size

        Returns
        -------
        list
            Generated truth table
        """

        return list(product((True, False), repeat=dimension))

    def test_dummy_treeview_class_with_wrong_model(self):
        """ Check geode_gem.widgets.view.CommonView.__init__ method
        """

        with self.assertRaises(TypeError):
            DummyGeodeGtkTreeView(None)

        with self.assertRaises(TypeError):
            DummyGeodeGtkTreeView(Gtk.Label())

    def test_dummy_treeview_class_with_correct_model(self):
        """ Check geode_gem.widgets.view.CommonView.__init__ method
        """

        DummyGeodeGtkTreeView(Gtk.ListStore())
        DummyGeodeGtkTreeView(Gtk.TreeStore())

    def test_dummy_treeview_class(self):
        """ Check geode_gem.widgets.view.CommonView.__init__ method
        """

        model = Gtk.ListStore(str)

        view = DummyGeodeGtkTreeView(model)
        self.assertEqual(view.get_model(), model)
        self.assertEqual(view.inner_model, view.list_model)
        self.assertFalse(view.is_filterable)
        self.assertNotEqual(view.inner_model, view.filtered_model)
        self.assertFalse(view.is_sorterable)
        self.assertNotEqual(view.inner_model, view.sorted_model)
        self.assertIsNone(view.sorting_column)
        self.assertEqual(view.sorting_order, Gtk.SortType.ASCENDING)
        self.assertIsNone(view.sort_func)
        self.assertIsNone(view.visible_func)

    def test_dummy_treeview_class_with_kwargs(self):
        """ Check geode_gem.widgets.view.CommonView.__init__ method
        """

        model = Gtk.ListStore(str, int)

        view = DummyGeodeGtkTreeView(model,
                                     filterable=False,
                                     sorterable=True,
                                     sorting_column=0,
                                     sort_func=print)
        self.assertFalse(view.is_filterable)
        self.assertTrue(view.is_sorterable)
        self.assertEqual(view.sorting_column, 0)
        self.assertIsNotNone(view.sort_func)

    def test_append_rows(self):
        """ Check geode_gem.widgets.view.CommonView.append method
        """

        view, treeiters = self.generate_treeview_and_treeiters()

        with self.assertRaises(ValueError):
            view.append([0])

        with self.assertRaises(TypeError):
            view.append([0, 0])

        model = view.get_model()
        self.assertIsNotNone(model)
        self.assertEqual(len(model), 10)
        self.assertEqual(model.get_string_from_iter(treeiters[0]), '0')
        self.assertEqual(model.get_string_from_iter(treeiters[5]), '5')
        self.assertEqual(model.get_value(treeiters[5], 0), 5)
        self.assertEqual(model.get_value(treeiters[7], 1), "index-7")

    def test_remove_rows(self):
        """ Check geode_gem.widgets.view.CommonView.remove method
        """

        view, treeiters = self.generate_treeview_and_treeiters()

        with self.assertRaises(TypeError):
            view.remove(None)

        view.remove(treeiters[0])
        view.remove(treeiters[5])
        view.remove(treeiters[6])

        model = view.get_model()
        self.assertEqual(len(model), 7)

        with self.assertRaises(TypeError):
            model.get_value(treeiters[5], 0)

    def test_clear_rows(self):
        """ Check geode_gem.widgets.view.CommonView.clear method
        """

        view = DummyGeodeGtkTreeView(Gtk.ListStore(int))
        for index in range(0, 10):
            view.append([index])

        self.assertEqual(len(view.get_model()), 10)
        view.clear()
        self.assertEqual(len(view.get_model()), 0)

    def test_convert_child_iter_to_iter(self):
        """ Check geode_gem.widgets.view.CommonView.convert_child_iter_to_iter
            method
        """

        for filterable, sorterable in self.generate_truth_table():
            view, treeiters = self.generate_treeview_and_treeiters(
                filterable=filterable, sorterable=sorterable)

            for index in range(len(treeiters)):
                self.check_converted_treeiter(view, treeiters[index], index)

    def test_get_path_from_treeiter(self):
        """ Check geode_gem.widgets.view.CommonView.get_path_from_treeiter
            method
        """

        for filterable, sorterable in self.generate_truth_table():
            view, treeiters = self.generate_treeview_and_treeiters(
                filterable=filterable, sorterable=sorterable)

            for index in range(len(treeiters)):
                treepath = view.get_path_from_treeiter(treeiters[index])
                self.assertIsNotNone(treepath)
                self.assertEqual(treepath.to_string(), str(index))

    def test_get_value(self):
        """ Check geode_gem.widgets.view.CommonView.get_value method
        """

        for filterable, sorterable in self.generate_truth_table():
            view, treeiters = self.generate_treeview_and_treeiters(
                filterable=filterable, sorterable=sorterable)

            for index in range(len(treeiters)):
                self.assertEqual(view.get_value(treeiters[index], 0), index)
                self.assertEqual(
                    view.get_value(treeiters[index], 1), f"index-{index}")

    def test_set_value(self):
        """ Check geode_gem.widgets.view.CommonView.set_value method
        """

        for filterable, sorterable in self.generate_truth_table():
            view, treeiters = self.generate_treeview_and_treeiters(
                filterable=filterable, sorterable=sorterable)

            for index in range(len(treeiters)):
                treeiter, value = treeiters[index], f"new-value-for-{index}"
                view.set_value(treeiter, 1, value)
                self.assertEqual(view.get_value(treeiter, 1), value)

    def test_refilter(self):
        """ Check geode_gem.widgets.view.CommonView.refilter method
        """

        with self.assertRaises(ValueError):
            DummyGeodeGtkTreeView(Gtk.ListStore(str)).refilter()

        with self.assertRaises(ValueError):
            DummyGeodeGtkTreeView(
                Gtk.ListStore(str), sorterable=True).refilter()
